/**
 * @property {number} circuit_id
 * @property {number} command
 * @property {Buffer} payload
 * @class
 */
class Cell {
  constructor (circuit_id, command, payload) {
    this.circuit_id = circuit_id;
    this.command = command;
    this.payload = payload;
  }

  is_recognized() {
    return this.payload[1] === 0 && this.payload[2] === 0;
  }
  /**
   * @param {number} protocol_version
   * @returns {Buffer}
   */
  get_bytes(protocol_version) {
    let cell_bytes;

    if (is_variable_length_cell_command(this.command))
    {
      cell_bytes = Buffer.alloc(
        //
        // circuit id.
        //
        (protocol_version < 4 ? 2 : 4) +

        //
        // Cell command.
        //
        1 +

        //
        // payload size (16 bits).
        //
        2 +

        //
        // payload.
        //
        this.payload.length);
    }
    else
    {
      cell_bytes = Buffer.alloc(Cell.size);
    }

    //
    // tor-spec.txt
    // 5.1.1.
    //
    // In link protocol 3 or lower, CircIDs are 2 bytes long;
    // in protocol 4 or higher, CircIDs are 4 bytes long.
    //

    let n = 0;
    if (protocol_version < 4)
    {
      n = cell_bytes.writeUInt16BE(this.circuit_id, n);
      //cell_buffer.write<circuit_id_v3_type>(static_cast<circuit_id_v3_type>(_circuit_id));
    }
    else
    {
      n = cell_bytes.writeInt32BE(this.circuit_id, n);
    }

    n = cell_bytes.writeUInt8(this.command, n);

    if (is_variable_length_cell_command(this.command))
    {
      n = cell_bytes.writeUInt16BE(this.payload.length, n);
    }

    this.payload.copy(cell_bytes, n);

    return cell_bytes;
  }
}
Cell.size = 514;
Cell.payload_size = 509;

function is_variable_length_cell_command(command) {
  return (command === Cell.commands.versions) || (command >= Cell.commands.vpadding);
}
Cell.is_variable_length_cell_command = is_variable_length_cell_command;

Cell.commands = Object.freeze({
  //
  // Cell commands.
  //
  padding: 0,
  create: 1,
  created: 2,
  relay: 3,
  destroy: 4,
  create_fast: 5,
  created_fast: 6,
  versions: 7,
  netinfo: 8,
  relay_early: 9,
  create2: 10,
  created2: 11,

  //
  // relay commands.
  //
  relay_begin: 1,
  relay_data: 2,
  relay_end: 3,
  relay_connected: 4,
  relay_sendme: 5,
  relay_extend: 6,
  relay_extended: 7,
  relay_truncate: 8,
  relay_truncated: 9,
  relay_drop: 10,
  relay_resolve: 11,
  relay_resolved: 12,
  relay_begin_dir: 13,
  relay_extend2: 14,
  relay_extended2: 15,

  //
  // rendezvous commands.
  //
  relay_command_establish_intro: 32,
  relay_command_establish_rendezvous: 33,
  relay_command_introduce1: 34,
  relay_command_introduce2: 35,
  relay_command_rendezvous1: 36,
  relay_command_rendezvous2: 37,
  relay_command_intro_established: 38,
  relay_command_rendezvous_established: 39,
  relay_command_introduce_ack: 40,

  //
  // variable-length Cell commands.
  //
  vpadding: 128,
  certs: 129,
  auth_challenge: 130,
  authenticate: 131,
  authorize: 132,
});
Cell.commands_lookup = Object.keys(Cell.commands).filter(x => x === 'relay_early' || !x.startsWith('relay_'))
  .reduce((a, c) => ({...a, [Cell.commands[c]]: c}), {});
Cell.relay_commands_lookup = Object.keys(Cell.commands).filter(x => x.startsWith('relay_'))
  .reduce((a, c) => ({...a, [Cell.commands[c]]: c}), {});

Cell.destroy_reasons = Object.freeze({
  none                           :  0, // no reason given
  protocol                       :  1, // tor protocol violation
  internal                       :  2, // internal error
  requested                      :  3, // a client sent a TRUNCATE command
  hibernating                    :  4, // not currently operating; trying to save bandwidth
  resource_limit                 :  5, // out of memory, sockets, or circuit IDs
  connection_failed              :  6, // unable to reach relay
  onion_router_identity          :  7, // connected to relay, but its OR identity was not as expected
  onion_router_connection_closed :  8, // the OR connection that was carrying this circuit died
  finished                       :  9, // the circuit has expired for being dirty or old
  timeout                        : 10, // circuit construction took too long
  destroyed                      : 11, // the circuit was destroyed w/o client TRUNCATE
  no_such_service                : 12, // request for unknown hidden service
});
Cell.destroy_reasons_lookup = Object.keys(Cell.destroy_reasons)
  .reduce((a,c) => ({...a, [Cell.destroy_reasons[c]]: c}), {});

module.exports = Cell;
