const HiddenServiceConnector = require('./HiddenServiceConnector');
const CircuitNode = require('./CircuitNode');
const Cell = require('./Cell')
const RelayCell = require('./RelayCell')
const Stream = require('./TorStream')
const {sha1} = require('../utils/crypto')
const { parseIp } = require('../utils/net')
const { defer } = require('../utils/time')

const States = Object.freeze({
  none:0,
  creating:1,
  extending:2,
  connecting:3,
  ready:4,
  destroyed:5,

  rendezvous_establishing:6,
  rendezvous_established:7,
  rendezvous_introducing:8,
  rendezvous_introduced:9,
  rendezvous_completing:10,
  rendezvous_completed:11,
});

class Circuit {
  static _next_circuit_id = 1;
  static _next_stream_id = 1;

  /** @type CircuitNode[] */
  _node_list = [];
  /** @type Socket */
  _socket = null;
  _circuit_id;
  _waits = {};
  /** @type CircuitNode */
  _extend_node = null;
  _stream_map = {};

  constructor (socket, logger) {
    this._socket = socket;
    this._circuit_id = (Circuit._next_circuit_id++) | 0x80000000;
    this._logger = logger;
  }

  get circuit_id() {
    return this._circuit_id;
  }

  get node_list() {
    return this._node_list;
  }

  get tor_socket() {
    return this._socket;
  }

  get state() { return this._state };

  set state(state) {
    this._state = state;
    const waiters = this._waits[state];
    if (waiters) {
      let d;
      while(d = waiters.pop()) {
        clearTimeout(d.timeout);
        d.resolve(true);
      }
    }
  }

  close_streams() {
    //
    // destroy each stream in this circuit.
    //
    for(const stream_id in this._stream_map)
    {
      //
      // this call removes the stream from our stream map.
      //
      this.send_relay_end_cell(this._stream_map[stream_id]);
    }

    this._node_list = [];

    //
    // signal destroy.
    //
    this.state = States.destroyed;
  }

  destroy() {
    if (this.state === States.destroyed)
    {
      return;
    }

    this._logger.debug("Circuit.destroy()");

    if (this._extend_node)
    {
      this._extend_node = null;
    }

    this.close_streams();
    this._socket.remove_circuit(this);
  }

  handle_cell(cell) {
    switch(cell.command){
      case Cell.commands.created2:
        this._on_created2(cell);
        break;
      case Cell.commands.destroy:
        this._on_destroy(cell);
        break;
      case Cell.commands.relay:
        const relay_cell = this._decrypt(cell);
        if (!relay_cell.node) {
          this._logger.warn(`could not decrypt cell for circuit ${this._circuit_id}. destroying`);
          this.destroy();
          return;
        }

        this._logger.debug(
          "Circuit.handle_cell() [circuit: %i%s, stream: %i, command: %s, relay_command: %s, payload_size: %i]",
          relay_cell.circuit_id & 0x7FFFFFFF,
          (relay_cell.circuit_id & 0x80000000 ? " (MSB set)" : ""),
          relay_cell.stream_id,
          Cell.commands_lookup[relay_cell.command],
          Cell.relay_commands_lookup[relay_cell.relay_command],
          relay_cell.relay_payload.length);

        switch(relay_cell.relay_command) {
          case Cell.commands.relay_extended2:
            this._on_relay_extended2(relay_cell);
            break;

          case Cell.commands.relay_truncated:
            this._on_relay_truncated_cell(relay_cell);
            break;

          case Cell.commands.relay_end:
            this._on_relay_end_cell(relay_cell);
            break;

          case Cell.commands.relay_connected:
            this._on_relay_connected_cell(relay_cell);
            break;

          case Cell.commands.relay_extended:
            this._on_relay_extended_cell(relay_cell);
            break;

          case Cell.commands.relay_data:
            this._on_relay_data_cell(relay_cell);
            break;

          case Cell.commands.relay_sendme:
            this._on_relay_sendme_cell(relay_cell);
            break;

          case Cell.commands.relay_command_rendezvous2:
            this._on_relay_extended_cell(relay_cell);
            this.state = States.rendezvous_completed;
            break;

          case Cell.commands.relay_command_rendezvous_established:
            this.state = States.rendezvous_established;
            break;

          case Cell.commands.relay_command_introduce_ack:
            this.state = States.rendezvous_introduced;
            break;

          default:
            this._logger.warn(
              "Circuit._handle_cell() !! unhandled relay cell [ relay_command: %s ]",
              Cell.relay_commands_lookup[relay_cell.relay_command]);
            break;
        }
        break;
      default:
        this._logger.warn(`unhandled cell for circuit ${this._circuit_id}`, cell);
    }
  }

  _on_relay_data_cell(cell) {
    this._logger.debug('Circuit._on_relay_data_cell():', cell.relay_payload);
    //
    // decrement deliver window on circuit node.
    //
    cell.node.decrement_deliver_window();
    if (cell.node.consider_sending_sendme()) {
      this.send_relay_sendme_cell(null);
    }

    const stream = this.get_stream_by_id(cell.stream_id);
    if (stream) {
      stream.append_to_recv_buffer(cell.relay_payload);

      //
      // decrement window on stream.
      //
      stream.decrement_deliver_window();
      if (stream.consider_sending_sendme()) {
        this.send_relay_sendme_cell(stream);
      }
    }
  }

  _on_relay_sendme_cell(cell) {
    if (cell.stream_id === 0) {
      cell.node.increment_package_window();
    } else {
      const stream = this.get_stream_by_id(cell.stream_id);
      if (stream) {
        stream.increment_package_window();
      }
    }
  }

  _on_relay_connected_cell(cell) {
    const stream = this.get_stream_by_id(cell.stream_id)
    if (stream) {
      stream.state = Stream.states.ready;
    }

    this.state = States.ready;
  }

  _on_relay_truncated_cell(cell) {
    //
    // tor-spec.txt
    // 5.4.
    //
    // To tear down part of a circuit, the OP may send a RELAY_TRUNCATE cell
    // signaling a given OR (Stream ID zero).  That OR sends a DESTROY
    // cell to the next node in the circuit, and replies to the OP with a
    // RELAY_TRUNCATED cell.
    //
    // [Note: If an OR receives a TRUNCATE cell and it has any RELAY cells
    // still queued on the circuit for the next node it will drop them
    // without sending them.  This is not considered conformant behavior,
    // but it probably won't get fixed until a later version of Tor.  Thus,
    // clients SHOULD NOT send a TRUNCATE cell to a node running any current
    // version of Tor if a) they have sent relay cells through that node,
    // and b) they aren't sure whether those cells have been sent on yet.]
    //
    // When an unrecoverable error occurs along one connection in a
    // circuit, the nodes on either side of the connection should, if they
    // are able, act as follows:  the node closer to the OP should send a
    // RELAY_TRUNCATED cell towards the OP; the node farther from the OP
    // should send a DESTROY cell down the circuit.
    //

    this._logger.error("circuit::handle_relay_truncated_cell() destroying circuit");
    this.destroy();
  }

  _on_relay_end_cell(cell) {
    const stream = this.get_stream_by_id(cell.stream_id);
    if (stream) {
      this._logger.debug("circuit::handle_relay_end_cell() [stream: %i, reason: %s]", cell.stream_id, Cell.destroy_reasons_lookup[cell.relay_payload[0]]);

      stream.state = Stream.states.destroyed;
      delete this._stream_map[cell.stream_id];
    }
  }

  _on_destroy(cell) {
    const reason = cell.payload.readUInt8(0);
    this._logger.info('Circuit._on_destroy:', Cell.destroy_reasons_lookup[reason]);
    this.destroy();
  }

  _on_relay_extended_cell(cell) {
    //TODO
  }

  _on_relay_extended2(cell) {
    //
    // The payload of an EXTENDED2 cell is the same as the payload of a
    // CREATED2 cell
    //

    //
    // finish the handshake.
    //
    const handshake_data = cell.relay_payload.slice(2);
    this._extend_node.compute_shared_secret(handshake_data);

    if (this._extend_node.has_valid_crypto_state())
    {
      this._node_list.push(this._extend_node);

      //
      // we're ready here.
      //
      this._extend_node = null;
      this.state = States.ready;
    }
    else
    {
      this._logger.error(`circuit::handle_extended2_cell() extend node [ ${this._extend_node.onion_router.name} ] has invalid crypto state`);

      this.destroy();
    }
  }

  _on_created2(cell) {
    //
    // A CREATED2 cell contains:
    //   HLEN      (Server Handshake Data Len) [2 bytes]
    //   HDATA     (Server Handshake Data)     [HLEN bytes]
    //

    //
    // finish the handshake.
    //
    const handshake_data = cell.payload.slice(2);
    this._extend_node.compute_shared_secret(handshake_data);

    if (this._extend_node.has_valid_crypto_state())
    {
      this._node_list.push(this._extend_node);

      //
      // we're ready here.
      //
      this._extend_node = null;
      this.state = States.ready;
    }
    else
    {
      this._logger.error(`circuit._handle_cell created2 extend node [ ${this._extend_node.onion_router.name} ] has invalid crypto state`);

      this.destroy();
    }
  }

  create(first_or, handshake_type) {
    switch (handshake_type) {
      case 'ntor':
        return this._create_ntor(first_or);
      default:
        throw new Error("Not implemented");
    }
  }

  async _create_ntor(first_or) {
    await first_or.fetch_descriptor();
    if (!first_or.ntor_onion_key)
    {
      this._logger.warn(`circuit._create_ntor() [or: ${first_or.name} does not support NTOR handshake]`);
      return;
    }

    this._logger.debug(`circuit._create_ntor() [or: ${first_or.name}, state: creating]`);
    this.state = States.creating;
    this._extend_node = new CircuitNode(this, first_or, 'normal');

    const handshake_bytes = Buffer.alloc(2 + 2 + 84);

    handshake_bytes.writeUInt16BE(2, 0);  // ntor type
    handshake_bytes.writeUInt16BE(84, 2); // ntor onion skin length
    const onion_skin_ntor = this._extend_node.create_onion_skin_ntor();
    onion_skin_ntor.copy(handshake_bytes, 4);

    this._socket.send_cell(new Cell(
      this._circuit_id,
      Cell.commands.create2,
      handshake_bytes
    ));

    if (await this.wait_for_state(States.ready))
    {
      this._logger.debug(`circuit._create_ntor() [or: ${first_or.name}, state: created]`);
    }
    else
    {
      this._logger.error(`circuit._create_ntor() [or: ${first_or.name}, state: destroyed]`);
    }
  }

  async create_stream(host, port) {
    //
    // tor-spec.txt
    // 6.2.
    //
    // ADDRPORT [nul-terminated string]
    // FLAGS[4 bytes]
    //
    // ADDRPORT is made of ADDRESS | ':' | PORT | [00]
    //

    const host_port = `${host}:${port}`

    const relay_data_bytes = Buffer.alloc(host_port.length + 1);
    relay_data_bytes.write(host_port, 0, 'utf8');

    //
    // send RELAY_BEGIN cell.
    //
    const stream_id = Circuit._next_stream_id++;

    const stream = new Stream(stream_id, this, this._logger);
    this._stream_map[stream_id] = stream;

    this._logger.debug("circuit::create_stream() [url: %s, stream: %i, status: creating]", host_port, stream_id);
    this.state = States.connecting;

    this.send_relay_cell(stream_id, Cell.commands.relay_begin, relay_data_bytes);

    if (await this.wait_for_state(States.ready))
    {
      this._logger.debug("circuit::create_stream() [url: %s, stream: %i, status: created]", host_port, stream_id);
      return stream;
    }
    else
    {
      this._logger.error("circuit::create_stream() [is_ready() == false]");
    }
  }

  async create_onion_stream(onion, port) {
    const hidden_service_connector = new HiddenServiceConnector(this, onion);

    return await hidden_service_connector.connect()
      ? this.create_stream(onion, port)
      : null;
  }

  create_dir_stream() {
    //TODO
    throw new Error('Not implemented');
  }

  extend(or, handshake_type = 'ntor') {
    switch (handshake_type) {
      case 'ntor':
        return this._extend_ntor(or);
      default:
        throw new Error("Not implemented");
    }
  }

  async _extend_ntor(next_onion_router) {
    //
    // An EXTEND2 cell's relay payload contains:
    //     NSPEC      (Number of link specifiers)     [1 byte]
    //       NSPEC times:
    //         LSTYPE (Link specifier type)           [1 byte]
    //         LSLEN  (Link specifier length)         [1 byte]
    //         LSPEC  (Link specifier)                [LSLEN bytes]
    //     HTYPE      (Client Handshake Type)         [2 bytes]
    //     HLEN       (Client Handshake Data Len)     [2 bytes]
    //     HDATA      (Client Handshake Data)         [HLEN bytes]
    //
    // Link specifiers describe the next node in the circuit and how to
    // connect to it. Recognized specifiers are:
    //    [00] TLS-over-TCP, IPv4 address
    //         A four-byte IPv4 address plus two-byte ORPort
    //    [01] TLS-over-TCP, IPv6 address
    //         A sixteen-byte IPv6 address plus two-byte ORPort
    //    [02] Legacy identity
    //         A 20-byte SHA1 identity fingerprint. At most one may be listed.
    //

    await next_onion_router.fetch_descriptor();
    if (!next_onion_router.ntor_onion_key)
    {
      this._logger.warn(`circuit._extend_ntor() [or: ${next_onion_router.name} does not support NTOR handshake`);
      return;
    }

    const ipv4 = 0;
    const ipv6 = 1;
    const legacy_id = 2;

    this._logger.debug(`circuit::extend_ntor() [or: ${next_onion_router.name}, state: extending]`);
    this.state = States.extending;

    this._extend_node = new CircuitNode(this, next_onion_router, 'normal');
    const onion_skin = this._extend_node.create_onion_skin_ntor();

    const relay_payload_bytes = Buffer.alloc(
    1 +                                // NSPEC
    1 + 1 +  6 +                     // NSPEC IPv4 (4 bytes) + port (2 bytes)
    1 + 1 + 20 +                     // NSPEC identity_fingerprint (20 bytes)
    2 +                                // HTYPE
    2 +                                // HLEN
    84);                               // HDATA

    let n = relay_payload_bytes.writeUInt8(2, 0); // 2x NSPEC

    n = relay_payload_bytes.writeUInt8(ipv4, n);
    n = relay_payload_bytes.writeUInt8(6, n);
    const ip = parseIp(next_onion_router.ip);
    n = relay_payload_bytes.writeInt32BE(ip, n);
    n = relay_payload_bytes.writeUInt16BE(next_onion_router.or_port, n);

    n = relay_payload_bytes.writeUInt8(legacy_id, n);
    n = relay_payload_bytes.writeUInt8(20, n);
    n += next_onion_router.identity_fingerprint.copy(relay_payload_bytes, n);

    n = relay_payload_bytes.writeUInt16BE(2, n);
    n = relay_payload_bytes.writeUInt16BE(84, n);
    onion_skin.copy(relay_payload_bytes, n);

    this.send_relay_cell(
      0,
      Cell.commands.relay_extend2,
      relay_payload_bytes,
      //
      // clients MUST only send
      // EXTEND cells inside RELAY_EARLY cells
      //
      Cell.commands.relay_early,
      this._extend_node);

    if (await this.wait_for_state(States.ready))
    {
      this._logger.debug(`circuit::extend_ntor() [or: ${next_onion_router.name}, state: extended]`);
    }
    else
    {
      this._logger.error(`circuit::extend_ntor() [or: ${next_onion_router.name}, state: destroyed]`);
    }
  }

  /**
   * @returns {CircuitNode}
   */
  get_final_circuit_node() {
    return this._node_list[this._node_list.length - 1];
  }

  get_stream_by_id(stream_id) {
    return this._stream_map[stream_id];
  }

  send_relay_sendme_cell() {
    //TODO
  }

  send_relay_end_cell(stream) {
    this.send_relay_cell(
      stream.stream_id,
      Cell.commands.relay_end,
      Buffer.from([6])); // reason

    //
    // signal destroy.
    //
    stream.state = Stream.states.destroyed;

    delete this._stream_map[stream.stream_id];
  }

  send_relay_data_cell(stream, data) {
    for (let i = 0; i < data.length; i += Cell.payload_size)
    {
      const data_size = Math.min(data.length - i, Cell.payload_size);

      this.get_final_circuit_node().decrement_package_window();

      this.send_relay_cell(
        stream.stream_id,
        Cell.commands.relay_data,
        data.slice(i, i + data_size));
    }
  }

  send_relay_cell(stream_id, relay_command, payload, cell_command = Cell.commands.relay_early, node = null) {
    node = node ? node : this.get_final_circuit_node();

    if (this.get_stream_by_id(stream_id) == null && stream_id !== 0)
    {
      this._logger.warn("Circuit.send_relay_cell() attempt to send cell to non-existent stream-id:", stream_id);
      return;
    }

    this._logger.debug(
      "Circuit.send_relay_cell() [circuit: %i%s, stream: %i, command: %s, relay_command: %s]",
      this._circuit_id & 0x7FFFFFFF,
      ((this._circuit_id & 0x80000000) ? " (MSB set)" : ""),
      stream_id,
      Cell.commands_lookup[cell_command],
      Cell.relay_commands_lookup[relay_command],
      payload);

    this._socket.send_cell(this._encrypt(new RelayCell(
      this._circuit_id,
      cell_command,
      node,
      relay_command,
      stream_id,
      payload)));
  }

  /**
   * @param {RelayCell} cell
   * @return {Cell}
   * @private
   */
  _encrypt(cell) {
    for (let i = this._node_list.length - 1; i >= 0; i--) {
      this._node_list[i].encrypt_forward_cell(cell);
    }
    return cell;
  }

  /**
   * @param {Cell} cell
   * @return {RelayCell}
   * @private
   */
  _decrypt(cell) {
    for (const node of this._node_list) {
      if (node.decrypt_backward_cell(cell)) {
        return RelayCell.fromCell(node, cell);
      }
    }
    return new RelayCell();
  }

  wait_for_state(desired_state, timeout = 30000) {
    const d = defer();
    d.timeout = setTimeout(() => {
      this._logger.debug("Circuit.wait_for_state(): timed out");
      d.resolve(false);
    }, timeout);
    (this._waits[desired_state] = this._waits[desired_state] || []).push(d);
    return d.promise;
  }

  async rendezvous_establish(rendezvous_cookie) {
    //TODO
    throw new Error("Not implemented");
  }

  async rendezvous_introduce(rendezvous_circuit, rendezvous_cookie) {
    //mini_assert(rendezvous_cookie.get_size() == 20);

    const introduction_point = this.get_final_circuit_node().onion_router;
    const introducee = rendezvous_circuit.get_final_circuit_node.onion_router;

    this._logger.debug("circuit::rendezvous_introduce() [or: %s, state: introducing]", introduction_point.name);
    this.state = States.rendezvous_introducing;

    this._logger.debug("circuit::rendezvous_introduce() [or: %s, state: completing]", introduction_point.name);
    rendezvous_circuit.state = States.rendezvous_completing;

    //
    // payload of the RELAY_COMMAND_INTRODUCE1
    // command:
    //
    // PK_ID  Identifier for Bob's PK      [20 octets]
    // VER    Version byte: set to 2.        [1 octet]
    // IP     Rendezvous point's address    [4 octets]
    // PORT   Rendezvous point's OR port    [2 octets]
    // ID     Rendezvous point identity ID [20 octets]
    // KLEN   Length of onion key           [2 octets]
    // KEY    Rendezvous point onion key [KLEN octets]
    // RC     Rendezvous cookie            [20 octets]
    // g^x    Diffie-Hellman data, part 1 [128 octets]
    //

    //
    // compute PK_ID, aka hash of the service key.
    //
    const service_key_hash = sha1(introduction_point.service_key);

    //
    // create rest of the payload in separate buffer;
    // it will be encrypted.
    //
    const handshake_bytes = Buffer.alloc(
      1 +  // version
      4  +  // ip address
      2  +  // port
      20 +  // identity_fingerprint
      2  +  // onion key size
      32 +  // onion key
      20 +  // rendezvous cookie
      128); // DH

    //io::memory_stream handshake_stream(handshake_bytes);
    //io::stream_wrapper handshake_buffer(handshake_stream, endianness::big_endian);

    rendezvous_circuit._extend_node = new CircuitNode(this, introduction_point, circuit_node_type::introduction_point);

    handshake_buffer.write(static_cast<uint8_t>(2));
    handshake_buffer.write(swap_endianness(introducee->get_ip_address().to_int()));
    handshake_buffer.write(introducee->get_or_port());
    handshake_buffer.write(introducee->get_identity_fingerprint());
    handshake_buffer.write(static_cast<payload_size_type>(introducee->get_onion_key().get_size()));
    handshake_buffer.write(introducee->get_onion_key());
    handshake_buffer.write(rendezvous_cookie);
    handshake_buffer.write(rendezvous_circuit->_extend_node->get_key_agreement().get_public_key());

    const handshake_encrypted = hybrid_encryption::encrypt(
      handshake_bytes,
      introduction_point.service_key);

      //
      // compose the final payload.
      //
    const relay_payload_bytes = Buffer.concat([
      service_key_hash,
      handshake_encrypted
    ], service_key_hash.length + handshake_encrypted.length);

    //
    // send the cell.
    //
    this.send_relay_cell(
      0,
      Cell.commands.relay_command_introduce1,
      relay_payload_bytes);

    if (await this.wait_for_state(States.rendezvous_introduced))
    {
      this._logger.debug("circuit::rendezvous_introduce() [or: %s, state: introduced]", introduction_point.name);
    }
    else
    {
      this._logger.error("circuit::rendezvous_introduce() [or: %s, is_rendezvous_introduced() == false]", introduction_point.name);

      //
      // we cannot expect the rendezvous will be completed.
      //

      return;
    }

    if (await rendezvous_circuit.wait_for_state(States.rendezvous_completed))
    {
      this._logger.debug("circuit::rendezvous_introduce() [or: %s, state: completed]", introduction_point.name);
    }
    else
    {
      this._logger.error("circuit::rendezvous_introduce() [or: %s, is_rendezvous_completed() == false]", introduction_point.name);
    }
  }

  is_ready() {
    return this.state === States.ready;
  }

  is_rendezvous_established() {
    return this.state === States.rendezvous_established;
  }

  is_rendezvous_completed() {
    return this.state === States.rendezvous_completed;
  }

  is_rendezvous_introduced() {
    return this.state === States.rendezvous_introduced;
  }
}

module.exports = Circuit;
