const fs = require('fs');
const crypto = require('crypto');
const {get} = require('../utils/http');
const OR = require('./OnionRouter');
const OnionRouter = require('./OnionRouter')

function authority_onion_router(name, ip, or_port, dir_port) {
  return {
    name,
    ip,
    or_port,
    dir_port
  }
}
const authorities = [
  authority_onion_router("dizum",      "194.109.206.212", 443,  80),
  authority_onion_router("Serge",      "66.111.2.131",    9001, 9030),
  authority_onion_router("moria1",     "128.31.0.34",     9101, 9131),
  //authority_onion_router("tor26",      "86.59.21.38",     443,  80},
  authority_onion_router("bastet",     "204.13.164.118",  443,  80),
  //authority_onion_router("maatuska",   "171.25.193.9",    80,   443),
  authority_onion_router("dannenberg", "193.23.244.244",  443,  80),
  authority_onion_router("Faravahar",  "154.35.175.225",  443,  80),
  authority_onion_router("gabelmoo",   "131.188.40.189",  443,  80),
  authority_onion_router("longclaw",   "199.58.81.140",   443,  80),
]
const router_status_flags = [
  "Authority",
  "BadExit",
  "Exit",
  "Fast",
  "Guard",
  "HSDir",
  "Named",
  "NoEdConsensus",
  "Stable",
  "Running",
  "Unnamed",
  "Valid",
  "V2Dir"
];

class Consensus {
  _allowed_dir_flags = OR.fast | OR.valid | OR.running | OR.v2dir;
  _allowed_dir_ports = [];
  _max_try_count = 3;
  _onion_router_map = {};
  _valid_until = 0;

  constructor (logger) {
    this._logger = logger;
  }

  async fetch(cached_consensus_path) {
    let have_valid_consensus = false;
    let force_download = false;
    let consensus_content = '';

    //
    // if no path to the cached consensus file
    // was provided, we have to download it.
    //
    if (!cached_consensus_path || !fs.existsSync(cached_consensus_path))
    {
      force_download = true;
    }

    while (!have_valid_consensus)
    {
      consensus_content = force_download
        ? await this._download_from_random_router("/tor/status-vote/current/consensus", true)
        : fs.readFileSync(cached_consensus_path).toString();

      //
      // assume newly downloaded consensus as valid.
      //
      const reject_invalid = !force_download;
      this._parse_consensus(consensus_content, reject_invalid);

      //
      // consider force_download-ed consensus as valid.
      //
      have_valid_consensus = force_download || (this._valid_until >= Date.now());

      //
      // if the consensus is invalid, we have to download it anyway.
      //
      if (!have_valid_consensus)
      {
        force_download = true;
      }
    }

    this._logger.debug("loaded", Object.keys(this._onion_router_map).length, "routers");

    //
    // save the consensus content, if the path was provided.
    //
    if (force_download && cached_consensus_path)
    {
      fs.writeFileSync(cached_consensus_path, consensus_content);
    }
  }

  set_allowed_dir_ports(...ports) {
    this._allowed_dir_ports = [...ports];
  }

  get_onion_routers_by_criteria(criteria) {
    const candidates = [];
    for (const key in this._onion_router_map) {
      const router = this._onion_router_map[key];
      let allowed = true;
      if (criteria.dir_ports) {
        allowed = allowed && criteria.dir_ports.includes(router.dir_port);
      }
      if (criteria.or_ports) {
        allowed = allowed && criteria.or_ports.includes(router.or_port);
      }
      if (criteria.flags) {
        allowed = allowed && (router.flags & criteria.flags) === criteria.flags;
      }
      if (criteria.forbidden_routers) {
        allowed = allowed && !criteria.forbidden_routers.includes(router.identity_fingerprint.toString('hex'))
      }
      if (allowed) {
        candidates.push(router);
      }
    }
    return candidates;
  }

  get_random_onion_router_by_criteria(criteria) {
    const candidates = this.get_onion_routers_by_criteria(criteria);
    if (candidates.length === 0) {
      return null;
    }
    const idx = crypto.randomInt(candidates.length);
    return candidates[idx];
  }

  /**
   * @param {Buffer} identity_fingerprint
   * @return {Promise<string>}
   */
  get_onion_router_descriptor(identity_fingerprint) {
    return this._download_from_random_router('/tor/server/fp/' + identity_fingerprint.toString('hex'));
  }

  /**
   * @param {string} path
   * @param {boolean} only_authorities
   * @return {Promise<string>}
   * @private
   */
  async _download_from_random_router(path, only_authorities = false) {
    let try_count = 0;
    let result = '';

    do {
      result = await this._download_from_random_router_impl(path, only_authorities);
    } while(++try_count < this._max_try_count && !result);

    return result;
  }

  /**
   * @param {string} path
   * @param {boolean} only_authorities
   * @return {Promise<string>}
   * @private
   */
  _download_from_random_router_impl(path, only_authorities) {
    let ip;
    let port;

    //
    // if the onion router map is empty,
    // we're stuck to authorities anyway.
    //
    if (only_authorities || Object.keys(this._onion_router_map).length === 0)
    {
      const random_index = crypto.randomInt(authorities.length);
      const authority = authorities[random_index];

      ip = authority.ip;
      port = authority.dir_port;
    }
    else
    {
      const router = this.get_random_onion_router_by_criteria({
        dir_ports: this._allowed_dir_ports,
        flags: this._allowed_dir_flags
      });

      ip = router.ip;
      port = router.dir_port;
    }

    this._logger.debug(`Consensus._download_from_random_router_impl() [path: http://${ip}:${port}${path}]`);

    return get('http:', ip, port, path);
  }

  _parse_consensus(content) {
    this._onion_router_map = {};
    this._valid_until = 0;

    const lines = content.split('\n');
    let current_router;
    for (const line of lines) {
      const parts = line.split(' ');
      switch (parts[0]) {
        case 'valid-until':
          this._valid_until = new Date(`${parts[1]}T${parts[2]}Z`).getTime();
          break;
        case 'r':
          const identity_fingerprint = Buffer.from(parts[2], 'base64');
          const identity_fingerprint_hex = identity_fingerprint.toString('hex');
          const or_port = parseInt(parts[7], 10);
          const dir_port = parseInt(parts[8], 10);
          current_router = new OnionRouter(this, parts[1], parts[6], or_port, dir_port, identity_fingerprint);
          this._onion_router_map[identity_fingerprint_hex] = current_router;
          break;
        case 's':
          let flags = 0;
          for (const part of parts) {
            const idx = router_status_flags.indexOf(part);
            if (idx === -1) continue;
            flags |= 1 << idx;
          }
          current_router.flags = flags;
          break;
        case 'directory-footer':
          return;
      }
    }
  }

  /**
   * @param {Buffer} identity_fingerprint
   * @returns {OnionRouter}
   */
  get_onion_router_by_identity_fingerprint(identity_fingerprint) {
    return this._onion_router_map[identity_fingerprint.toString('hex')];
  }
}

module.exports = Consensus;
