const crypto = require('crypto');
const hkdf = require('futoin-hkdf');
const curve25519 = require('curve25519-js');
const { hmac_sha256 } = require('../utils/crypto')

const const_server = 'Server';
const const_protoid = 'ntor-curve25519-sha256-1';
const const_t_mac = 'ntor-curve25519-sha256-1:mac';
const const_t_verify = 'ntor-curve25519-sha256-1:verify';
const const_m_expand = 'ntor-curve25519-sha256-1:key_expand';
const const_t_key = 'ntor-curve25519-sha256-1:key_extract';

class KeyAgreementNtor {
  /** @type Buffer */
  _private_key = null;
  /** @type Buffer */
  _public_key = null;

  constructor (or) {
    this._onion_router = or;
    const kp = curve25519.generateKeyPair(crypto.randomBytes(32));
    this._private_key = Buffer.from(kp.private);
    this._public_key = Buffer.from(kp.public);
  }

  get public_key() {
    return this._public_key;
  }

  compute_shared_secret(handshake_data) {
    return this._compute_shared_secret(handshake_data.slice(0, 32), handshake_data.slice(32, 32 + 32));
  }

  _compute_shared_secret(other_public_key, verification_data) {
    //
    // 5.1.4. The "ntor" handshake
    //
    // In this section, define:
    //    H(x,t) as HMAC_SHA256 with message x and key t.
    //    H_LENGTH  = 32.
    //    ID_LENGTH = 20.
    //    G_LENGTH  = 32
    //    PROTOID   = "ntor-curve25519-sha256-1"
    //    t_mac     = PROTOID | ":mac"
    //    t_key     = PROTOID | ":key_extract"
    //    t_verify  = PROTOID | ":verify"
    //    MULT(a,b) = the multiplication of the curve25519 point 'a' by the
    //                scalar 'b'.
    //    G         = The preferred base point for curve25519 ([9])
    //    KEYGEN()  = The curve25519 key generation algorithm, returning
    //                a private/public keypair.
    //    m_expand  = PROTOID | ":key_expand"
    //    KEYID(A)  = A
    //
    // To perform the handshake, the client needs to know an identity key
    // digest for the server, and an ntor onion key (a curve25519 public
    // key) for that server. Call the ntor onion key "B".  The client
    // generates a temporary keypair:
    //     x,X = KEYGEN()
    // and generates a client-side handshake with contents:
    //     NODEID      Server identity digest  [ID_LENGTH bytes]
    //     KEYID       KEYID(B)                [H_LENGTH bytes]
    //     CLIENT_PK   X                       [G_LENGTH bytes]
    //
    // The server generates a keypair of y,Y = KEYGEN(), and uses its ntor
    // private key 'b' to compute:
    //
    //   secret_input = EXP(X,y) | EXP(X,b) | ID | B | X | Y | PROTOID
    //   KEY_SEED = H(secret_input, t_key)
    //   verify = H(secret_input, t_verify)
    //   auth_input = verify | ID | B | Y | X | PROTOID | "Server"
    //
    // The server's handshake reply is:
    //     SERVER_PK   Y                       [G_LENGTH bytes]
    //     AUTH        H(auth_input, t_mac)    [H_LENGTH bytes]
    //
    // The client then checks Y is in G^* [see NOTE below], and computes
    //
    //   secret_input = EXP(Y,x) | EXP(B,x) | ID | B | X | Y | PROTOID
    //   KEY_SEED = H(secret_input, t_key)
    //   verify = H(secret_input, t_verify)
    //   auth_input = verify | ID | B | Y | X | PROTOID | "Server"
    //
    // The client verifies that AUTH == H(auth_input, t_mac).
    //

    //0d f8 4c b3 97 82 67 a5 b4 56 29 cc 21 c3 6f 0e 9f 83 48 3f 97 d4 45 bf 79 2b 51 7e e5 91 c8 2e

    const shared_key1 = Buffer.from(curve25519.sharedKey(this._private_key, other_public_key));
    const shared_key2 = Buffer.from(curve25519.sharedKey(this._private_key, this._onion_router.ntor_onion_key));

    const secret_input = Buffer.concat([
      shared_key1,
      shared_key2,
      this._onion_router.identity_fingerprint,
      this._onion_router.ntor_onion_key,
      this._public_key,
      other_public_key,
      Buffer.from(const_protoid)
    ], 32 + 32 + 20 + 32 + 32 + 32 + const_protoid.length);

    const verify = hmac_sha256(const_t_verify, secret_input);

    const auth_input = Buffer.concat([
      verify,
      this._onion_router.identity_fingerprint,
      this._onion_router.ntor_onion_key,
      other_public_key,
      this._public_key,
      Buffer.from(const_protoid),
      Buffer.from(const_server)
    ], 32 + 20 + 32 + 32 + 32 + const_protoid.length + const_server.length);

    const computed_verification_data = hmac_sha256(const_t_mac, auth_input);

    if (verification_data.equals(computed_verification_data))
    {
      //
      // create key material.
      //
      return hkdf(secret_input, 92, {
        salt: const_t_key,
        info: const_m_expand,
        hash: 'sha256'
      });
    }

    return null;
  }
}
module.exports = KeyAgreementNtor;
