const net = require('net');
const tls = require('tls');
const Cell = require('./Cell');
const Circuit = require('./Circuit');
const { defer } = require('../utils/time')
const { wait } = require('../utils/time')
const { time } = require('../utils/time')
const { parseIp } = require('../utils/net')

const protocol_version_initial   = 3;
const protocol_version_preferred = 4;

const States = {
  closed: 0,
  connecting: 1,
  handshake_in_progress: 2,
  ready: 3
}

class Socket {
  /** @type TLSSocket */
  _socket = null;
  _onion_router = null;
  _protocol_version = protocol_version_initial;
  _circuit_map = {};
  _state = States.closed
  _waits = {};
  // recv
  _receive_payload_size = 0;
  /** @type Cell */
  _cell = null;

  constructor (logger) {
    this._logger = logger;
  }

  get onion_router() {
    return this._onion_router;
  }
  get state() {
    return this._state;
  }
  set state(state) {
    this._state = state;
    const waiters = this._waits[state];
    if (waiters) {
      let d;
      while(d = waiters.pop()) {
        clearTimeout(d.timeout);
        d.resolve(true);
      }
    }
  }

  async connect(or) {
    if (this.is_connected()) {
      this.close();
    }

    this.state = States.connecting;
    this._onion_router = or;
    this._logger.debug("Connecting to", or.name, "@", or.ip, or.or_port);
    this._socket = new tls.connect({
      host: or.ip,
      port: or.or_port,
      rejectUnauthorized: false
    });
    await new Promise((resolve, reject) => {
      this._socket.once('secureConnect', resolve);
      this._socket.once('error', reject);
    });
    this._socket.on('data', this._recv_cell.bind(this));

    if (!this.is_connected()) {
      this.state = States.closed;
      return;
    }

    this.state = States.handshake_in_progress;

    this._send_versions();

    await this.wait_for_state(States.ready);
  }

  async close() {
    this._socket.end();
    this._socket.unref();
    this.state = States.closed;
  }

  is_connected() {
    return this._socket && !this._socket.connecting;
  }

  /**
   * @param handshake
   * @returns {Promise<Circuit>}
   */
  async create_circuit(handshake = 'ntor') {
    if (this.state !== States.ready)
    {
      return null;
    }

    const new_circuit = new Circuit(this, this._logger);
    this._circuit_map[new_circuit.circuit_id] = new_circuit;
    await new_circuit.create(this._onion_router, handshake);

    //
    // should't we close the socket in the case of failure?
    //
    return new_circuit.node_list.length === 1
      ? new_circuit
      : null;
  }

  remove_circuit(circuit) {
    this._logger.debug("tor_socket::remove_circuit() [circuit: %i]", circuit.circuit_id & 0x7FFFFFFF);

    delete this._circuit_map[circuit.circuit_id];
  }

  async wait_for_state(desired_state, timeout = 30000) {
    const d = defer();
    d.timeout = setTimeout(() => {
      this._logger.debug("Socket.wait_for_state(): timed out");
      d.resolve(false);
    }, timeout);
    (this._waits[desired_state] = this._waits[desired_state] || []).push(d);
    return d.promise;
  }

  _recv_cell(chunk) {
    for (let i = 0; i < chunk.length;) {
      if (!this._cell) {
        let circuit_id;
        if (this._protocol_version < 4) {
          circuit_id = chunk.readUInt16BE(i);
          i += 2;
        } else {
          circuit_id = chunk.readInt32BE(i);
          i += 4;
        }
        let command = chunk.readUInt8(i);
        i++;
        let payload_size = Cell.payload_size;
        if (Cell.is_variable_length_cell_command(command)) {
          payload_size = chunk.readUInt16BE(i);
          i += 2;
        }
        this._cell = new Cell(circuit_id, command, Buffer.alloc(payload_size));
        this._receive_payload_left = payload_size;
      }
      const targetStart = this._cell.payload.length - this._receive_payload_left;
      const length = Math.min(chunk.length - i, this._receive_payload_left);
      const copied = chunk.copy(this._cell.payload, targetStart, i, i + length);
      this._receive_payload_left -= copied;
      i += copied;

      if (this._receive_payload_left === 0) {
        const cell = this._cell;
        this._cell = null;
        this._handle_cell(cell);
      }
    }
  }

  _handle_cell(cell) {
    this._logger.debug("Socket._handle_cell(): %i %s", cell.circuit_id & 0x7fffffff, Cell.commands_lookup[cell.command], cell.payload);
    if (cell.circuit_id) {
      return this._circuit_map[cell.circuit_id].handle_cell(cell);
    }
    switch(cell.command) {
      case Cell.commands.versions:
        this._on_recv_versions(cell);
        break;
      case Cell.commands.certs:
        //this._logger.debug("Socket recv certificates", cell);
        break;
      case Cell.commands.auth_challenge:
        //this._logger.debug("Socket recv auth challenge", cell);
        break;
      case Cell.commands.netinfo:
        this._on_recv_net_info(cell);
        this._send_net_info();
        this.state = States.ready;
        break;
      default:
        this._logger.debug("Cell unhandled");
        break;
    }
  }

  send_cell(cell) {
    if (!this.is_connected()) return;

    this._logger.debug("Socket.send_cell(): %i %s", (cell.circuit_id & 0x7fffffff), Cell.commands_lookup[cell.command], cell.payload);
    const bytes = cell.get_bytes(this._protocol_version);
    this._socket.write(bytes);
  }

  _send_versions() {
    this._logger.debug("Socket._send_versions()");
    this.send_cell(new Cell(0, Cell.commands.versions, Buffer.from([0, 4])));
  }

  _on_recv_versions(versions_cell) {
    this._logger.debug("Socket._recv_versions()");

    for (let i = 0; i < versions_cell.payload.length; i += 2)
    {
      const offered_version = versions_cell.payload.readUInt16BE(i);

      if (offered_version === protocol_version_preferred)
      {
        this._protocol_version = offered_version;
        this._logger.debug("switched to protocol v4");
        break;
      }
    }
  }

  _on_recv_net_info(net_info_cell) {
    this._logger.debug("Socket._recv_net_info()");
    //const cell = this._recv_cell();
    this._localAddress = "127.0.0.1";
    if (net_info_cell.payload.readUInt8(4) === 4 && net_info_cell.payload.readUInt8(5) === 4) {
      const addr = net_info_cell.payload.readUInt32BE(6);
      this._localAddress = [
          ((addr >> 24) & 0xff),
          ((addr >> 16) & 0xff),
          ((addr >> 8) & 0xff),
          (addr & 0xff)
        ].join('.')
      this._logger.debug("local addr =", this._localAddress);
    }
  }

  _send_net_info() {
    this._logger.debug("Socket._send_net_info()");

    const remote = parseIp(this._socket.remoteAddress);
    const local = parseIp(this._localAddress);
    const epoch = time();

    const net_info_bytes = Buffer.alloc(4 + 2 + 4 + 3 + 4);

    //
    // If version 2 or higher is negotiated, each party sends the other a
    // NETINFO Cell.  The Cell's payload is:
    //
    //  Timestamp              [4 bytes]
    //  Other OR's address     [variable]
    //  Number of addresses    [1 byte]
    //  This OR's addresses    [variable]
    //
    // Address is:
    //   Type   (1 octet)
    //   Length (1 octet)
    //   Value  (variable-width)
    //
    //  "Type" is one of:
    //    0x00 -- Hostname
    //    0x04 -- IPv4 address
    //    0x06 -- IPv6 address
    //    0xF0 -- Error, transient
    //    0xF1 -- Error, nontransient
    //

    let n = net_info_bytes.writeUInt32BE(epoch, 0);
    n = net_info_bytes.writeUInt8(4, n); // type
    n = net_info_bytes.writeUInt8(4, n); // length
    n = net_info_bytes.writeInt32BE(remote, n);
    n = net_info_bytes.writeUInt8(1, n); // number of addresses
    n = net_info_bytes.writeUInt8(4, n); // type
    n = net_info_bytes.writeUInt8(4, n); // length
    net_info_bytes.writeInt32BE(local, n);

    this.send_cell(new Cell(
      0,
      Cell.commands.netinfo,
      net_info_bytes));
  }
}

module.exports = Socket;
