const { Duplex } = require('stream');

const States = Object.freeze({
  none: 0,
  connecting: 1,
  ready: 2,
  destroyed: 3,
});

class TorStream extends Duplex {
  static states = States;
  static window_start = 500;
  static window_increment = 50;
  static window_max_unflushed = 10;

  _stream_id = 0;
  /** @type Circuit */
  _circuit = null;
  _deliver_window = TorStream.window_start;
  _package_window = TorStream.window_start;
  /** @type Buffer */
  _buffer = null;
  _state = States.connecting;
  _waits = {};

  /**
   * @param {number} stream_id
   * @param {Circuit} circuit
   * @param logger
   */
  constructor (stream_id, circuit, logger) {
    super();
    this._stream_id = stream_id;
    this._circuit = circuit;
    this._logger = logger;
  }

  get stream_id() { return this._stream_id; }
  get state() { return this._state; }
  set state(state) {
    this._state = state;
    if (state === States.destroyed) {
      for (const k in this._waits) {
        for (const wait of this._waits[k]) {
          wait.resolve(false);
        }
      }
      this._waits = {};
    }
  }

  close() {
    if (this.state === States.destroyed) {
      return;
    }
    this._circuit.send_relay_end_cell(this);
  }

  append_to_recv_buffer(data) {
    if (this._canRead) {
      this._canRead = this.push(data);
    } else {
      this._buffer = this._buffer
        ? Buffer.concat([this._buffer, data], this._buffer.length + data.length)
        : data;
    }
  }

  decrement_deliver_window() {
    this._deliver_window--;
  }

  consider_sending_sendme() {
    if (this._deliver_window > (TorStream.window_start - TorStream.window_increment)) {
      this._logger.debug("TorStream.consider_sending_sendme(): false");
      return false;
    }

    //
    // we're currently flushing immediatelly upon write,
    // therefore there is no need to check unflushed cell count,
    // because it's always 0.
    //
    // if (unflushed_cell_count >= window_max_unflushed)
    // {
    //   return false;
    // }
    //

    this._deliver_window += TorStream.window_increment;

    this._logger.debug("TorStream.consider_sending_sendme(): true");
    return true;
  }

  increment_package_window() {
    this._deliver_window += TorStream.window_increment;
  }

  get_stream() {
    return this;
  }

  _write(chunk, encoding, callback) {
    let err = null;
    try {
      this._circuit.send_relay_data_cell(this, chunk);
    } catch (error) {
      err = error;
    } finally {
      callback(err);
    }
  }

  _read(size) {
    if (this._buffer) {
      this._canRead = this.push(this._buffer);
      this._buffer = null;
    } else {
      this._canRead = true;
    }
  }
}
module.exports = TorStream;
